# -*- coding: utf-8 -*-
"""
    bromelia.etsi_3gpp.ts_129_272
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains Diameter AVP classes defined in ETSI TS 129 272.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from ..etsi_3gpp.ts_129_061 import X3gppChargingCharacteristicsAVP

from ..etsi_3gpp.ts_129_212 import PreEmptionCapabilityAVP
from ..etsi_3gpp.ts_129_212 import PreEmptionVulnerabilityAVP
from ..etsi_3gpp.ts_129_212 import PriorityLevelAVP
from ..etsi_3gpp.ts_129_212 import QosClassIdentifierAVP

from ..etsi_3gpp.ts_129_214 import MaxRequestedBandwidthDlAVP
from ..etsi_3gpp.ts_129_214 import MaxRequestedBandwidthUlAVP

from ..etsi_3gpp.ts_129_229 import VisitedNetworkIdentifierAVP

from ..etsi_3gpp.ts_129_329 import MsisdnAVP

from ..ietf.rfc5447 import Mip6AgentInfoAVP

from ...base import DiameterAVP
from ...constants.etsi_3gpp.ts_129_272 import *
from ...exceptions import AVPAttributeValueError
from ...types import *
from ...utils import decode_from_tbcd, encode_to_tbcd


class StnSrAVP(DiameterAVP, OctetStringType):
    """Implementation of STN-SR AVP in Section 7.3.39 of 
    ETSI TS 129 272 V15.4.0 (2018-07).

    The STN-SR AVP (AVP Code 1433) is of type OctetString.
    """
    code = STN_SR_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             StnSrAVP.code,
                             StnSrAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        OctetStringType.__init__(self, data=self.encode(data), vendor_id=VENDOR_ID_3GPP)


    def encode(self, data):
        if isinstance(data, int):
            return bytes.fromhex(encode_to_tbcd(data))

        elif isinstance(data, str):
            return bytes.fromhex(encode_to_tbcd(int(data)))

        elif isinstance(data, bytes):
            return data


    def decode(self):
        return decode_from_tbcd(self.data)


class AmbrAVP(DiameterAVP, GroupedType):
    """Implementation of AMBR AVP in Section 7.3.41 of 
    ETSI TS 129 272 V15.10.0 (2020-01).

    The AMBR AVP (AVP Code 1435) is of type Grouped.
    """
    code = AMBR_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    mandatory = {
                    "max_requested_bandwidth_ul": MaxRequestedBandwidthUlAVP,
                    "max_requested_bandwidth_dl": MaxRequestedBandwidthDlAVP,
    }
    optionals = {
                    # "extended_max_requested_bw_ul": ExtendedMaxRequestedBwUlAVP,
                    # "extended_max_requested_bw_dl": ExtendedMaxRequestedBwDlAVP,
    }

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             AmbrAVP.code,
                             AmbrAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        GroupedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class SubscriberStatusAVP(DiameterAVP, EnumeratedType):
    """Implementation of Subscriber-Status AVP in Section 7.3.29 of 
    ETSI TS 129 272 V15.4.0 (2018-07).

    The Subscriber-Status AVP (AVP code 1424) is of type Enumerated.
    """
    code = SUBSCRIBER_STATUS_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    values = [
                 SUBSCRIBER_STATUS_SERVICE_GRANTED,
                 SUBSCRIBER_STATUS_OPERATOR_DETERMINED_BARRING,
     ]

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             SubscriberStatusAVP.code,
                             SubscriberStatusAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        EnumeratedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class ContextIdentifierAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Context-Identifier AVP in Section 7.3.27
    of ETSI TS 129 272 V15.10.0 (2020-01).

    The Context-Identifier AVP (AVP Code 1423) is of type Unsigned32.
    """
    code = CONTEXT_IDENTIFIER_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             ContextIdentifierAVP.code,
                             ContextIdentifierAVP.vendor_id)
        DiameterAVP.set_vendor_id_bit(self, True)
        Unsigned32Type.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class PdnTypeAVP(DiameterAVP, EnumeratedType):
    """Implementation of PDN-Type AVP in Section 7.3.62 of
    ETSI TS 129 272 V15.10.0 (2020-01).

    The PDN-Type AVP (AVP Code 1456) is of type Enumerated.
    """
    code = PDN_TYPE_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    values = [
                PDN_TYPE_IPV4,
                PDN_TYPE_IPV6,
                PDN_TYPE_IPV4V6,
                PDN_TYPE_IPV4_OR_IPV6
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             PdnTypeAVP.code,
                             PdnTypeAVP.vendor_id)
        DiameterAVP.set_vendor_id_bit(self, True)    
        EnumeratedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class ServiceSelectionAVP(DiameterAVP, UTF8StringType):
    """Implementation of Service-Selection AVP in Section 7.3.36
    of ETSI TS 129 272 V15.10.0 (2020-01).

    The Service-Selection AVP (AVP Code 493) is of type UTF8String.
    """
    code = SERVICE_SELECTION_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             ServiceSelectionAVP.code,
                             ServiceSelectionAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        UTF8StringType.__init__(self, data=data)


class AllocationRetentionPriorityAVP(DiameterAVP, GroupedType):
    """Implementation of Allocation-Retention-Priority AVP in Section 7.3.40 
    of ETSI TS 129 272 V15.10.0 (2020-01).

    The Allocation-Retention-Priority AVP (AVP Code 1034) is of type Grouped.
    """
    code = ALLOCATION_RETENTION_PRIORITY_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    mandatory = {
                    "priority_level": PriorityLevelAVP
    }
    optionals = {
                    "pre_emption_capability": PreEmptionCapabilityAVP,
                    "pre_emption_vulnerability": PreEmptionVulnerabilityAVP

    }

    def __init__(self, data):
        DiameterAVP.__init__(self,
                             AllocationRetentionPriorityAVP.code,
                             AllocationRetentionPriorityAVP.vendor_id)
        DiameterAVP.set_vendor_id_bit(self, True)
        GroupedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class EpsSubscribedQosProfileAVP(DiameterAVP, GroupedType):
    """Implementation of EPS-Subscribed-QoS-Profile AVP in Section 7.3.37 
    of ETSI TS 129 272 V15.10.0 (2020-01)

    The EPS-Subscribed-QoS-Profile AVP (AVP Code 1431) is of type Grouped.
    """
    code = EPS_SUBSCRIBED_QOS_PROFILE_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    mandatory = {
                    "qos_class_identifier": QosClassIdentifierAVP,
                    "allocation_retention_priority": AllocationRetentionPriorityAVP
    }
    optionals = {}

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             EpsSubscribedQosProfileAVP.code,
                             EpsSubscribedQosProfileAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        GroupedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class VplmnDynamicAddressAllowedAVP(DiameterAVP, EnumeratedType):
    """Implementation of VPLMN-Dynamic-Address-Allowed AVP in Section 7.3.38 
    of ETSI TS 129 272 V15.10.0 (2020-01).

    The VPLMN-Dynamic-Address-Allowed AVP (AVP Code 1432) is of type 
    Enumerated.
    """
    code = VPLMN_DYNAMIC_ADDRESS_ALLOWED_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    values = [
                VPLMN_DYNAMIC_ADDRESS_ALLOWED_NOT_ALLOWED,
                VPLMN_DYNAMIC_ADDRESS_ALLOWED_ALLOWED
    ]


    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             VplmnDynamicAddressAllowedAVP.code,
                             VplmnDynamicAddressAllowedAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        EnumeratedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class PdnGwAllocationTypeAVP(DiameterAVP, EnumeratedType):
    """Implementation of PDN-GW-Allocation-Type AVP in Section 7.3.44 of 
    ETSI TS 129 272 V15.10.0 (2020-01).

    The PDN-GW-Allocation-Type AVP (AVP Code 1438) is of type Enumerated.
    """
    code = PDN_GW_ALLOCATION_TYPE_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    values = [
                PDN_GW_ALLOCATION_TYPE_STATIC,
                PDN_GW_ALLOCATION_TYPE_DYNAMIC
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             PdnGwAllocationTypeAVP.code,
                             PdnGwAllocationTypeAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        EnumeratedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class ApnConfigurationAVP(DiameterAVP, GroupedType):
    """Implementation of APN-Configuration AVP in Section 7.3.35 of
    ETSI TS 129 272 V15.4.0 (2018-07).

    The APN-Configuration AVP (AVP Code 1430) is of type Grouped.
    """
    code = APN_CONFIGURATION_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    mandatory = {
                    "context_identifier": ContextIdentifierAVP,
                    "pdn_type": PdnTypeAVP,
                    "service_selection": ServiceSelectionAVP,
    }
    optionals = {
                    # "served_party_ip_address": ServedPartyIdAddressAVP,
                    "eps_subscribed_qos_profile": EpsSubscribedQosProfileAVP,
                    "vplmn_dynamic_address_allowed": VplmnDynamicAddressAllowedAVP,
                    "mip6_agent_info": Mip6AgentInfoAVP,
                    "visited_network_identifier": VisitedNetworkIdentifierAVP,
                    "pdn_gw_allocation_type": PdnGwAllocationTypeAVP,
                    "x3gpp_charging_characteristics": X3gppChargingCharacteristicsAVP,
                    "ambr": AmbrAVP,
                    # "specific_apn_info": SpecificApnInfoAVP,
                    # "apn_oi_replacement": ApnOiReplacementAVP,
                    # "sipto_permission": SiptoPermissionAVP,
                    # "lipa_permission": LipaPermissionAVP,
                    # "restoration_priority": RestorationPriorityAVP,
                    # "sipto_local_network_permission": SiptoLocalNetworkPermissionAVP,
                    # "wlan_offloadability": WlanOffloadabilityAVP,
                    # "non_ip_pdn_type_indicator": NonIpPdnTypeIndicatorAVP,
                    # "non_ip_data_delivery_mechanism": NonIpDataDeliveryMechanismAVP,
                    # "scef_id": ScefIdAVP,
                    # "scef_realm": ScefRealmAVP,
                    # "preferred_data_mode": PreferredDataModeAVP,
                    # "pdn_connectivity_continuity": PdnConnectivityContinuityAVP,
                    # "rds_indicator": RdsIndicatorAVP,
    }

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             ApnConfigurationAVP.code,
                             ApnConfigurationAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        GroupedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class AllApnConfigurationsIncludedIndicatorAVP(DiameterAVP, EnumeratedType):
    """Implementation of All-APN-Configurations-Included-Indicator AVP in 
    Section 7.3.33 of ETSI TS 129 272 V15.4.0 (2018-07).

    The All-APN-Configurations-Included-Indicator AVP (AVP Code 1428) is of 
    type Enumerated.
    """
    code = ALL_APN_CONFIGURATIONS_INCLUDED_INDICATOR_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    values = [
                ALL_APN_CONFIGURATIONS_INCLUDED,
                MODIFIED_ADDED_APN_CONFIGURATIONS_INCLUDED,
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             AllApnConfigurationsIncludedIndicatorAVP.code,
                             AllApnConfigurationsIncludedIndicatorAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)    
        EnumeratedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class ApnConfigurationProfileAVP(DiameterAVP, GroupedType):
    """Implementation of APN-Configuration-Profile AVP in Section 7.3.34 of
    ETSI TS 129 272 V15.4.0 (2018-07).

    The APN-Configuration-Profile AVP (AVP Code 1429) is of type Grouped.
    """
    code = APN_CONFIGURATION_PROFILE_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    mandatory = {
                    "context_identifier": ContextIdentifierAVP,
                    "all_apn_configurations_included_indicator": AllApnConfigurationsIncludedIndicatorAVP,
                    "apn_configuration": ApnConfigurationAVP,
    }
    optionals = {
                    # "additional_context_identifier": AdditionalContextIdentifierAVP,
    }

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             ApnConfigurationProfileAVP.code,
                             ApnConfigurationProfileAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        GroupedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class UeUsageTypeAVP(DiameterAVP, Unsigned32Type):
    """Implementation of UE-Usage-Type AVP in Section 7.3.202 of 
    ETSI TS 129 272 V15.4.0 (2018-07).

    The UE-Usage-Type AVP (AVP Code 1680) is of type Unsigned32.
    """
    code = UE_USAGE_TYPE_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             UeUsageTypeAVP.code,
                             UeUsageTypeAVP.vendor_id)
        DiameterAVP.set_vendor_id_bit(self, True)
        Unsigned32Type.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class OperatorDeterminedBarringAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Operator-Determined-Barring AVP in Section 7.3.30 of 
    ETSI TS 129 272 V15.10.0 (2020-01).

    The Operator-Determined-Barring AVP (AVP Code 1425) is of type Unsigned32.
    """
    code = OPERATOR_DETERMINED_BARRING_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data=convert_to_4_bytes(0)):
        DiameterAVP.__init__(self, 
                             OperatorDeterminedBarringAVP.code,
                             OperatorDeterminedBarringAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        Unsigned32Type.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class SubscriptionDataAVP(DiameterAVP, GroupedType):
    """Implementation of Subscription-Data AVP in Section 7.3.2 of 
    ETSI TS 129 272 V15.4.0 (2018-07).

    The Subscription-Data AVP (AVP Code 1400) is of type Grouped.
    """
    code = SUBSCRIPTION_DATA_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    optionals = {
                    "subscriber_status": SubscriberStatusAVP,
                    "msisdn": MsisdnAVP,
                    # "a_msisdn": AMsisdnAVP,
                    "stn_sr": StnSrAVP,
                    # "ics_indicator": IcsIndicatorAVP,
                    # "network_access_mode": NetworkAccessModeAVP,
                    "operator_determined_barring": OperatorDeterminedBarringAVP,
                    # "hplmn_odb": HplmnOdbAVP,
                    # "regional_subscription_zone_code": RegionalSubscriptionZoneCodeAVP,
                    # "access_restriction_data": AccessRestrictionDataAVP,
                    # "apn_oi_replacement": ApnOiReplacementAVP,
                    # "lcs_info": LcsInfoAVP,
                    # "teleservice_list": TeleserviceListAVP,
                    # "call_barring_info": CallBarringInfoAVP,
                    "x3pp_charging_characteristics": X3gppChargingCharacteristicsAVP,
                    "ambr": AmbrAVP,
                    "apn_configuration_profile": ApnConfigurationProfileAVP,
                    # "rat_frequency_selection_priority_id": RatFrequencySelectionPriorityIdAVP,
                    # "trace_data": TraceDataAVP,
                    # "gprs_subscription_data": GprsSubscriptionDataAVP,
                    # "mps_priority": MpsPriorityAVP,
                    # "vplmn_lipa_allowed": VplmnLipaAllowedAVP,
                    # "relay_node_indicator": RelayNodeIndicatorAVP,
                    # "mdt_user_consent": MdtUserConsentAVP,
                    # "subscribed_vsrvcc": SubscribedVsrvccAVP,
                    # "prose_subscription_data": ProseSubscriptionDataAVP,
                    # "subscription_data_flags": SubscriptionDataFlagsAVP,
                    # "adjacent_access_restriction_data": AdjacentAccessRestrictionDataAVP,
                    # "dl_buffering_suggested_packet_count": DlBufferingSuggestedPacketCountAVP,
                    # "imsi_group_id": ImsiGroupIdAVP,
                    "ue_usage_type": UeUsageTypeAVP,
                    # "aese_communication_pattern": AeseCommunicationPatternAVP,
                    # "monitoring_event_configuration": MonitoringEventConfigurationAVP,
                    # "emergency_info": EmergencyInfoAVP,
                    # "v2x_subscription_data": V2xSubscriptionDataAVP,
                    # "edrx_cycle_length": EdrxCycleLengthAVP,
                    # "external_identifier": ExternalIdentifierAVP,
                    # "active_time": ActiveTimeAVP,
                    # "service_gap_time": ServiceGapTimeAVP,
                    # "broadcast_location_assistance_data_types": BroadcastLocationAssistanceDataTypesAVP,
                    # "aerial_ue_subscription_information": AerialUeSubscriptionInformationAVP,
    }

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             SubscriptionDataAVP.code,
                             SubscriptionDataAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        GroupedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class ImeiAVP(DiameterAVP, UTF8StringType):
    """Implementation of IMEI AVP in Section 7.3.4 of
    ETSI TS 129 272 V15.10.0 (2020-01).

    The IMEI AVP (AVP Code 1402) is of type UTF8String.
    """
    code = IMEI_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             ImeiAVP.code,
                             ImeiAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        UTF8StringType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class NumberOfRequestedVectorsAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Number-Of-Requested-Vectors AVP in Section 7.3.14 of 
    ETSI TS 129 272 V15.10.0 (2020-01).

    The Number-Of-Requested-Vectors AVP (AVP Code 1410) is of type Unsigned32.
    """
    code = NUMBER_OF_REQUESTED_VECTORS_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             NumberOfRequestedVectorsAVP.code,
                             NumberOfRequestedVectorsAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        Unsigned32Type.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class ReSynchronizationInfoAVP(DiameterAVP, OctetStringType):
    """Implementation of Re-Synchronization-Info AVP in Section 7.3.15 of 
    ETSI TS 129 272 V15.10.0 (2020-01).

    The Re-Synchronization-Info AVP (AVP Code 1411) is of type OctetString.
    """
    code = RE_SYNCHRONIZATION_INFO_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             ReSynchronizationInfoAVP.code,
                             ReSynchronizationInfoAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        OctetStringType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class ImmediateResponsePreferredAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Immediate-Response-Preferred AVP in Section 7.3.16 of 
    ETSI TS 129 272 V15.10.0 (2020-01).

    The Immediate-Response-Preferred AVP (AVP Code 1412) is of type Unsigned32.
    """
    code = IMMEDIATE_RESPONSE_PREFERRED_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             ImmediateResponsePreferredAVP.code,
                             ImmediateResponsePreferredAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        Unsigned32Type.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class RequestedEutranAuthenticationInfoAVP(DiameterAVP, GroupedType):
    """Implementation of Requested-EUTRAN-Authentication-Info AVP in 
    Section 7.3.11 of ETSI TS 129 272 V15.10.0 (2020-01).

    The Requested-EUTRAN-Authentication-Info AVP (AVP Code 1408) is of type 
    Grouped.
    """
    code = REQUESTED_EUTRAN_AUTHENTICATION_INFO_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    mandatory = {}
    optionals = {
                    "number_of_requested_vectors": NumberOfRequestedVectorsAVP,
                    "immediate_response_preferred": ImmediateResponsePreferredAVP,
                    "re_synchronization_info": ReSynchronizationInfoAVP,
    }

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             RequestedEutranAuthenticationInfoAVP.code,
                             RequestedEutranAuthenticationInfoAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        GroupedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class SoftwareVersionAVP(DiameterAVP, UTF8StringType):
    """Implementation of Software-Version AVP in Section 7.3.5 of
    ETSI TS 129 272 V15.10.0 (2020-01).

    The Software-Version AVP (AVP Code 1403) is of type UTF8String.
    """
    code = SOFTWARE_VERSION_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             SoftwareVersionAVP.code,
                             SoftwareVersionAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        UTF8StringType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class TerminalInformationAVP(DiameterAVP, GroupedType):
    """Implementation of Terminal-Information AVP in Section 5.3.14 of
    ETSI TS 129 272 V15.10.0 (2020-01).

    The Terminal-Information AVP (AVP Code 1401) is of type Grouped.
    """
    code = TERMINAL_INFORMATION_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    mandatory = {}
    optionals = {
                     "imei": ImeiAVP,
                    # "x3gpp2_meid": x3gpp2MeidAVP,
                    "software_version": SoftwareVersionAVP       
    }

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             TerminalInformationAVP.code,
                             TerminalInformationAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        GroupedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class EquipmentStatusAVP(DiameterAVP, EnumeratedType):
    """Implementation of Equipment-Status AVP in Section 7.3.51 of 
    ETSI TS 129 272 V15.4.0 (2018-07).

    The Equipment-Status AVP (AVP Code 1445) is of type Enumerated.
    """
    code = EQUIPMENT_STATUS_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    values = [
                EQUIPMENT_STATUS_WHITELISTED,
                EQUIPMENT_STATUS_BLACKLISTED,
                EQUIPMENT_STATUS_GREYLISTED
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             EquipmentStatusAVP.code,
                             EquipmentStatusAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        EnumeratedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class UlrFlagsAVP(DiameterAVP, Unsigned32Type):
    """Implementation of ULR-Flags AVP in Section 7.3.7 of 
    ETSI TS 129 272 V15.10.0 (2020-01).

    The ULR-Flags AVP (AVP Code 1405) is of type Unsigned32.
    """
    code = ULR_FLAGS_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             UlrFlagsAVP.code,
                             UlrFlagsAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        Unsigned32Type.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class UlaFlagsAVP(DiameterAVP, Unsigned32Type):
    """Implementation of ULA-Flags AVP in Section 7.3.8 of 
    ETSI TS 129 272 V15.4.0 (2018-07).

    The ULA-Flags AVP (AVP Code 1406) is of type Unsigned32.
    """
    code = ULA_FLAGS_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             UlaFlagsAVP.code,
                             UlaFlagsAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        Unsigned32Type.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class AirFlagsAVP(DiameterAVP, Unsigned32Type):
    """Implementation of AIR-Flags AVP in Section 7.3.201 of 
    ETSI TS 129 272 V15.4.0 (2018-07).

    The AIR-Flags AVP (AVP Code 1679) is of type Unsigned32.
    """
    code = AIR_FLAGS_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             AirFlagsAVP.code,
                             AirFlagsAVP.vendor_id)
        DiameterAVP.set_vendor_id_bit(self, True)
        Unsigned32Type.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class NorFlagsAVP(DiameterAVP, Unsigned32Type):
    """Implementation of NOR-Flags AVP in Section 7.3.49 of 
    ETSI TS 129 272 V15.4.0 (2018-07).

    The NOR-Flags AVP (AVP Code 1443) is of type Unsigned32.
    """
    code = NOR_FLAGS_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             NorFlagsAVP.code,
                             NorFlagsAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        Unsigned32Type.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class PurFlagsAVP(DiameterAVP, Unsigned32Type):
    """Implementation of PUR-Flags AVP in Section 7.3.149 of 
    ETSI TS 129 272 V15.4.0 (2018-07).

    The PUR-Flags AVP (AVP Code 1635) is of type Unsigned32.
    """
    code = PUR_FLAGS_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             PurFlagsAVP.code,
                             PurFlagsAVP.vendor_id)
        DiameterAVP.set_vendor_id_bit(self, True)
        Unsigned32Type.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class PuaFlagsAVP(DiameterAVP, Unsigned32Type):
    """Implementation of PUA-Flags AVP in Section 7.3.49 of 
    ETSI TS 129 272 V15.4.0 (2018-07).

    The PUA-Flags AVP (AVP Code 1442) is of type Unsigned32.
    """
    code = PUA_FLAGS_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             PuaFlagsAVP.code,
                             PuaFlagsAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        Unsigned32Type.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class AlertReasonAVP(DiameterAVP, EnumeratedType):
    """Implementation of Alert-Reason AVP in Section 7.3.83 of 
    ETSI TS 129 272 V15.4.0 (2018-07).

    The Alert-Reason AVP (AVP Code 1434) is of type Enumerated.
    """
    code = ALERT_REASON_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    values = [
                ALERT_REASON_UE_PRESET,
                ALERT_REASON_UE_MEMORY_AVAILABLE
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             AlertReasonAVP.code,
                             AlertReasonAVP.vendor_id)
        DiameterAVP.set_vendor_id_bit(self, True)
        EnumeratedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class ErrorDiagnosticAVP(DiameterAVP, EnumeratedType):
    """Implementation of Error-Diagnostic AVP in Section 7.3.128 of
    ETSI TS 129 272 V15.4.0 (2018-07).

    The Error-Diagnostic AVP (AVP Code 1614) is of type Enumerated.
    """
    code = ERROR_DIAGNOSTIC_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    values = [
                ERROR_DIAGNOSTIC_GPRS_DATA_SUBSCRIBED,
                ERROR_DIAGNOSTIC_NO_GPRS_DATA_SUBSCRIBED,
                ERROR_DIAGNOSTIC_ODB_ALL_APN,
                ERROR_DIAGNOSTIC_ODB_HPLMN_APN,
                ERROR_DIAGNOSTIC_ODB_VPLMN_APN
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             ErrorDiagnosticAVP.code,
                             ErrorDiagnosticAVP.vendor_id)
        DiameterAVP.set_vendor_id_bit(self, True)
        EnumeratedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class VisitedPlmnIdAVP(DiameterAVP, OctetStringType):
    """Implementation of Visited-PLMN-Id AVP in Section 7.3.9 of 
    ETSI TS 129 272 V15.10.0 (2020-01).

    The Visited-PLMN-Id AVP (AVP Code 1407) is of type OctetString.
    """
    code = VISITED_PLMN_ID_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             VisitedPlmnIdAVP.code,
                             VisitedPlmnIdAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        OctetStringType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class ItemNumberAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Item-Number AVP in Section 7.3.23 of 
    ETSI TS 129 272 V15.4.0 (2018-07).

    The Item-Number AVP (AVP Code 1419) is of type Unsigned32.
    """
    code = ITEM_NUMBER_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             ItemNumberAVP.code,
                             ItemNumberAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        Unsigned32Type.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class RandAVP(DiameterAVP, OctetStringType):
    """Implementation of RAND AVP in Section 7.3.53 of 
    ETSI TS 129 272 V15.4.0 (2018-07).

    The RAND AVP (AVP Code 1447) is of type OctetString.
    """
    code = RAND_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             RandAVP.code,
                             RandAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        OctetStringType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class XresAVP(DiameterAVP, OctetStringType):
    """Implementation of XRES AVP in Section 7.3.54 of 
    ETSI TS 129 272 V15.4.0 (2018-07).

    The XRES AVP (AVP Code 1448) is of type OctetString.
    """
    code = XRES_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             XresAVP.code,
                             XresAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        OctetStringType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class AutnAVP(DiameterAVP, OctetStringType):
    """Implementation of AUTN AVP in Section 7.3.55 of 
    ETSI TS 129 272 V15.4.0 (2018-07).

    The AUTN AVP (AVP Code 1449) is of type OctetString.
    """
    code = AUTN_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             AutnAVP.code,
                             AutnAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        OctetStringType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class KasmeAVP(DiameterAVP, OctetStringType):
    """Implementation of KASME AVP in Section 7.3.56 of 
    ETSI TS 129 272 V15.4.0 (2018-07).

    The KASME AVP (AVP Code 1450) is of type OctetString.
    """
    code = KASME_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             KasmeAVP.code,
                             KasmeAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        OctetStringType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class EUtranVectorAVP(DiameterAVP, GroupedType):
    """Implementation of E-UTRAN-Vector AVP in Section 7.3.18 of
    ETSI TS 129 272 V15.4.0 (2018-07).

    The E-UTRAN-Vector AVP (AVP Code 1414) is of type Grouped.
    """
    code = E_UTRAN_VECTOR_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    mandatory = {}
    optionals = {
                     "item_number": ItemNumberAVP,
                     "rand": RandAVP,
                     "xres": XresAVP,
                     "autn": AutnAVP,
                     "kasme": KasmeAVP,
    }

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             EUtranVectorAVP.code,
                             EUtranVectorAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        GroupedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class AuthenticationInfoAVP(DiameterAVP, GroupedType):
    """Implementation of Authentication-Info AVP in Section 7.3.17 of
    ETSI TS 129 272 V15.4.0 (2018-07).

    The Authentication-Info AVP (AVP Code 1413) is of type Grouped.
    """
    code = AUTHENTICATION_INFO_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    mandatory = {}
    optionals = {
                     "e_utran_vector": EUtranVectorAVP,
                    #  "utran_vector": UtranVectorAVP,
                    #  "geran_vector": GeranVectorAVP,
    }

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             AuthenticationInfoAVP.code,
                             AuthenticationInfoAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        GroupedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class HomogeneousSupportOfImsVoiceOverPsSessionsAVP(DiameterAVP, EnumeratedType):
    """Implementation of Homogeneous-Support-of-IMS-Voice-Over-PS-Sessions AVP 
    in Section 7.3.107 of ETSI TS 129 272 V15.4.0 (2018-07).

    The Homogeneous-Support-of-IMS-Voice-Over-PS-Sessions AVP (AVP Code 1493) 
    is of type Enumerated.
    """
    code = HOMOGENEOUS_SUPPORT_OF_IMS_VOICE_OVER_PS_SESSION_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    values = [
                IMS_VOICE_OVER_PS_NOT_SUPPORTED,
                IMS_VOICE_OVER_PS_SUPPORTED
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             HomogeneousSupportOfImsVoiceOverPsSessionsAVP.code,
                             HomogeneousSupportOfImsVoiceOverPsSessionsAVP.vendor_id)
        DiameterAVP.set_vendor_id_bit(self, True)
        EnumeratedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class UeSrvccCapabilityAVP(DiameterAVP, EnumeratedType):
    """Implementation of UE-SRVCC-Capability AVP in Section 7.3.130 of 
    ETSI TS 129 272 V15.10.0 (2020-01).

    The UE-SRVCC-Capability AVP (AVP Code 1615) is of type Enumerated.
    """
    code = UE_SRVCC_CAPABILITY_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    values = [
                UE_SRVCC_NOT_SUPPORTED,
                UE_SRVCC_SUPPORTED
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             UeSrvccCapabilityAVP.code,
                             UeSrvccCapabilityAVP.vendor_id)
        DiameterAVP.set_vendor_id_bit(self, True)
        EnumeratedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class SupportedServicesAVP(DiameterAVP, GroupedType):
    """Implementation of Supported-Services AVP in Section 7.3.199 of 
    ETSI TS 129 272 V15.10.0 (2020-01).

    The Supported-Services AVP (AVP Code 3143) is of type Grouped.
    """
    code = SUPPORTED_SERVICES_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             SupportedServicesAVP.code,
                             SupportedServicesAVP.vendor_id)
        DiameterAVP.set_vendor_id_bit(self, True)
        GroupedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class SupportedMonitoringEventsAVP(DiameterAVP, Unsigned64Type):
    """Implementation of Supported-Monitoring-Events AVP in Section 7.3.200 of 
    ETSI TS 129 272 V15.10.0 (2020-01).

    The Supported-Monitoring-Events AVP (AVP Code 3144) is of type Unsigned64.
    """
    code = SUPPORTED_MONITORING_EVENTS_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             SupportedMonitoringEventsAVP.code,
                             SupportedMonitoringEventsAVP.vendor_id)
        DiameterAVP.set_vendor_id_bit(self, True)
        Unsigned64Type.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class CancellationTypeAVP(DiameterAVP, EnumeratedType):
    """Implementation of Cancellation-Type AVP in Section 7.3.24 of 
    ETSI TS 129 272 V15.4.0 (2018-07).

    The Cancellation-Type AVP (AVP Code 1420) is of type Enumerated.
    """
    code = CANCELLATION_TYPE_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    values = [
                CANCELLATION_TYPE_MME_UPDATE_PROCEDURE,
                CANCELLATION_TYPE_SGSN_UPDATE_PROCEDURE,
                CANCELLATION_TYPE_SUBSCRIPTION_WITHDRAWAL,
                CANCELLATION_TYPE_UPDATE_PROCEDURE_IWF,
                CANCELLATION_TYPE_INITIAL_ATTACH_PROCEDURE
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             CancellationTypeAVP.code,
                             CancellationTypeAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        EnumeratedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class ClrFlagsAVP(DiameterAVP, Unsigned32Type):
    """Implementation of CLR-Flags AVP in Section 7.3.152 of 
    ETSI TS 129 272 V15.4.0 (2018-07).

    The CLR-Flags AVP (AVP Code 1638) is of type Unsigned32.
    """
    code = CLR_FLAGS_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             ClrFlagsAVP.code,
                             ClrFlagsAVP.vendor_id)
        DiameterAVP.set_vendor_id_bit(self, True)
        Unsigned32Type.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)



# -*- coding: utf-8 -*-
"""
    bromelia.constants.ietf.rfc4006
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~    
    
    This module contains constants defined in IETF RFC 4006.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from ..._internal_utils import convert_to_4_bytes


#: Diameter AVPs
SUBSCRIPTION_ID_AVP_CODE = convert_to_4_bytes(443)
SUBSCRIPTION_ID_DATA_AVP_CODE = convert_to_4_bytes(444)
SUBSCRIPTION_ID_TYPE_AVP_CODE = convert_to_4_bytes(450)

#: List of Subscription-Id-Type values.
#: For more information, please refer to Section 8.47 of IETF RFC 4006.
END_USER_E164 = convert_to_4_bytes(0)
END_USER_IMSI = convert_to_4_bytes(1)
END_USER_SIP_URI = convert_to_4_bytes(2)
END_USER_NAI = convert_to_4_bytes(3)
END_USER_PRIVATE = convert_to_4_bytes(4)



# -*- coding: utf-8 -*-
"""
    bromelia.types
    ~~~~~~~~~~~~~~

    This module contains the base structures that power the
    bromelia.avps module.

    Defines the AVP library that are used to create Diameter messages.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

import abc
import datetime
import ipaddress
import re

from ._internal_utils import avp_look_up
from ._internal_utils import convert_to_3_bytes
from ._internal_utils import convert_to_4_bytes
from ._internal_utils import convert_to_8_bytes
from .base import DiameterAVP
from .base import loader
from .constants import *
from .exceptions import AVPAttributeValueError
from .exceptions import DataTypeError
from .exceptions import DiameterAvpError
from .exceptions import DiameterTypeError


class BaseDataType(abc.ABC):
    @classmethod
    def load(cls, avp):
        return cls(avp.data)


class OctetStringType(BaseDataType):
    @abc.abstractmethod
    def __init__(self, data, vendor_id=None):
        self.parser_data(data)

        if isinstance(vendor_id, bytes):
            if len(vendor_id) != 4:
                raise DataTypeError("Invalid vendor_id format for "\
                                    "OctetStringType. It MUST be 4 bytes long")

            self._length = convert_to_3_bytes(len(self.data) + 
                                                    AVP_HEADER_LENGTH_LONGER)
            self._vendor_id = vendor_id

        elif vendor_id is None:
            self._length = convert_to_3_bytes(len(self.data) + 
                                                    AVP_HEADER_LENGTH)
            self._vendor_id = None

        self._padding = self.set_padding()


    def parser_data(self, data):
        if isinstance(data, str):
            self._data = data.encode("utf-8")

        elif isinstance(data, bytes):
            self._data = data

        else:
            raise DataTypeError("OctetStringType MUST have data argument "\
                                "of 'bytes' or 'str'")


    def set_padding(self):
        mod = len(self.data) % 4
        if mod != 0:
            return bytes(4 - mod)
        else:
            return bytes(0)


class Integer32Type(BaseDataType):
    @abc.abstractmethod
    def __init__(self, data, vendor_id=None):
        self.parser_data(data)

        if isinstance(vendor_id, bytes):
            if len(vendor_id) != 4:
                raise DataTypeError("Integer32Type MUST have vendor_id "\
                                    "argument of 'bytes'")

            self._length = convert_to_3_bytes(16)
            self._vendor_id = vendor_id

        elif vendor_id is None:
            self._length = convert_to_3_bytes(12)
            self._vendor_id = None                        
    

    def parser_data(self, data):
        if len(data) != 4:
            raise DataTypeError("Integer32Type MUST have data argument "\
                                "of 'bytes' with 32-bit signed value")

        self._data = data


class Unsigned32Type(BaseDataType):
    @abc.abstractmethod
    def __init__(self, data, vendor_id=None):
        self.parser_data(data)

        if isinstance(vendor_id, bytes):
            if len(vendor_id) != 4:
                raise DataTypeError("Unsigned32Type MUST have vendor_id "\
                                    "argument of 'bytes'")

            self._length = convert_to_3_bytes(16)
            self._vendor_id = vendor_id

        elif vendor_id is None:
            self._length = convert_to_3_bytes(12)
            self._vendor_id = None                        


    def parser_data(self, data):
        if isinstance(data, int):
            self._data = convert_to_4_bytes(data)


        elif isinstance(data, bytes):
            if len(data) != 4:
                raise DataTypeError("Unsigned32Type MUST have data argument "\
                                    "of 'bytes' with 32-bit signed value")

            self._data = data


    def is_bit_set(self, bit):
        if 0 <= bit < 8:
            return self.data[3] & (2 ** bit) !=0
        elif 8 <= bit < 16:
            return self.data[2] & (2 ** (bit % 8)) !=0
        elif 16 <= bit < 24:
            return self.data[1] & (2 ** (bit % 8)) !=0
        elif 24 <= bit < 32:
            return self.data[0] & (2 ** (bit % 8)) !=0
        else:
            raise DiameterTypeError("Bit index out of range")


    def set_bit(self, bit):
        if self.is_bit_set(bit):
            raise DiameterTypeError("Unable to set an already set bit")

        if 0 <= bit < 8:
            data = self.data[3] | 2 ** bit
            self.data = bytes(bytearray([self.data[0], self.data[1], self.data[2], data]))
        elif 8 <= bit < 16:
            data = self.data[2] | 2 ** (bit % 8) 
            self.data = bytes(bytearray([self.data[0], self.data[1], data, self.data[3]]))
        elif 16 <= bit < 24:
            data = self.data[1] | 2 ** (bit % 8)
            self.data = bytes(bytearray([self.data[0], data, self.data[2], self.data[3]]))
        elif 24 <= bit < 32:
            data = self.data[0] | 2 ** (bit % 8)
            self.data = bytes(bytearray([data, self.data[1], self.data[2], self.data[3]]))

        return self.data


    def unset_bit(self, bit):
        if not self.is_bit_set(bit):
            raise DiameterTypeError("Unable to unset an already unset bit")

        if 0 <= bit < 8:
            data = self.data[3] ^ 2 ** bit
            self.data = bytes(bytearray([self.data[0], self.data[1], self.data[2], data]))
        elif 8 <= bit < 16:
            data = self.data[2] ^ 2 ** (bit % 8) 
            self.data = bytes(bytearray([self.data[0], self.data[1], data, self.data[3]]))
        elif 16 <= bit < 24:
            data = self.data[1] ^ 2 ** (bit % 8)
            self.data = bytes(bytearray([self.data[0], data, self.data[2], self.data[3]]))
        elif 24 <= bit < 32:
            data = self.data[0] ^ 2 ** (bit % 8)
            self.data = bytes(bytearray([data, self.data[1], self.data[2], self.data[3]]))

        return self.data


class Unsigned64Type(BaseDataType):
    @abc.abstractmethod
    def __init__(self, data, vendor_id=None):
        self.parser_data(data)

        if isinstance(vendor_id, bytes):
            if len(vendor_id) != 4:
                raise DataTypeError("Unsigned64Type MUST have vendor_id "\
                                    "argument of 'bytes'")

            self._length = convert_to_3_bytes(20)
            self._vendor_id = vendor_id

        elif vendor_id is None:
            self._length = convert_to_3_bytes(16)
            self._vendor_id = None                        


    def parser_data(self, data):
        if isinstance(data, int):
            self._data = convert_to_8_bytes(data)

        elif isinstance(data, bytes):
            if len(data) != 8:
                raise DataTypeError("Unsigned64Type MUST have data argument "\
                                    "of 'bytes' with 64-bit unsigned value")

            self._data = data


class GroupedType(BaseDataType):
    mandatory = {}
    optionals = {}

    @abc.abstractmethod
    def __init__(self, data, vendor_id=None):
        if isinstance(data, bytes):
            self._data = data
            self.avps = DiameterAVP.load(data)

        elif isinstance(data, list):
            self._data = b""
            self.avps = data
            for avp in data:
                if not isinstance(avp, DiameterAVP):
                    raise DataTypeError("GroupedType MUST have data "\
                                        "argument of 'list' with "\
                                        "DiameterAVP objects")
                
        else:
            raise DataTypeError("GroupedType MUST have data argument "\
                                "of 'list' with DiameterAVP objects")

        if self.mandatory:
            mandatory_avps_codes = [avp.code for avp in self.mandatory.values()]
            data_avp_codes = [avp.code for avp in self.avps]
            
            if not set(mandatory_avps_codes).issubset(data_avp_codes):
                raise AVPAttributeValueError("missing mandatory avp", 
                                              DIAMETER_MISSING_AVP)
        
        #: TO-DO: maybe create a custom condtional method to check specific
        #: rules for a given AVP. 

        if isinstance(vendor_id, bytes):
            if len(vendor_id) != 4:
                raise DataTypeError("GroupedType MUST have vendor_id "\
                                    "argument of 'bytes'")

            self._vendor_id = vendor_id
    
            if isinstance(data, bytes):
                _length = AVP_HEADER_LENGTH_LONGER
                self._length = convert_to_3_bytes(len(data) + _length)

        elif vendor_id is None:
            self._vendor_id = None

            if isinstance(data, bytes):
                _length = AVP_HEADER_LENGTH
                self._length = convert_to_3_bytes(len(data) + _length)


    @classmethod
    def load(cls, avp):
        self.append(avp)
        return cls(avp)


    @property
    def avps(self):
        return tuple(self._avps)


    @avps.setter
    def avps(self, value):
        if not isinstance(value, list):
            raise DiameterAvpError(f"only list allowed. Cannot append a "\
                                   f"data type of '{type(value)}'")
    
        self.cleanup()

        if len(value) == 1:
            self.append(value[0])
        elif len(value) > 1:
            self.extend(value)


    def __getitem__(self, idx):
        return self._avps[idx]


    def __setitem__(self, idx, value):
        self._avps[idx] = value


    def append(self, avp):
        if not isinstance(avp, DiameterAVP):
            raise DiameterAvpError(f"cannot append a data type of "\
                                   f"'{type(avp)}'")

        avp_name = avp_look_up(avp)
        if avp_name == "Unknown":
            avp_name = loader.get_avp_class_name(avp)

        _name = avp_name.replace("-", "_").lower()
        avp_key = f"{_name}_avp"


        if avp_key in self.__dict__:
            index = 0
            for key in self.__dict__.keys():
                if avp_key in key:
                    index += 1
            avp_key = f"{avp_key}__{index}"

        self._avps.append(avp)
        self.__dict__.update({avp_key: avp})

        self._data += avp.dump()


    def extend(self, avps):
        for avp in avps:
            self.append(avp)


    def has_avp(self, avp_key):
        if not isinstance(avp_key, str):
            raise DiameterAvpError("`avp_key` must be str")

        if not self.avps:
            return False
        elif avp_key in self.__dict__:
            return True
        return False


    def pop(self, avp_key):
        if not self.avps:
            raise DiameterAvpError("`avps` attribute is empty. There is "\
                                   "no DiameterAVP object to be removed")

        try:
            item = self.__dict__[avp_key]
        except KeyError:
            raise DiameterAvpError(f"`{avp_key}` key not defined")

        self._avps.remove(item)
        self.__dict__.pop(avp_key, None)

        self._data = b""
        for avp in self.avps:
            self._data += avp.dump()


    def update_key(self, old_avp_key, new_avp_key):
        if not self.has_avp(old_avp_key):
            raise DiameterAvpError(f"`{old_avp_key}` key not defined")
            
        if self.has_avp(new_avp_key):
            raise DiameterAvpError(f"`{new_avp_key}` key already defined")

        self.__dict__[new_avp_key] = self.__dict__.pop(old_avp_key)


    def cleanup(self):
        self._data = b""
        self._avps = list()

        avps_keys = list()
        for avp_key in self.__dict__.keys():
            if "_avp" in avp_key and avp_key != "_avps":
                avps_keys.append(avp_key)
                    
        for avp_key in avps_keys:
            self.__dict__.pop(avp_key, None)


    #: TO-DO: refresh does not work properly, it needs to refresh all data, 
    #: specialy if avp is of grouped type.
    def refresh(self):
        if self.is_vendor_id():
            avp_length = AVP_HEADER_LENGTH_LONGER
        else:
            avp_length = AVP_HEADER_LENGTH

        real_length = 0

        for avp in self.avps:
            avp_length += len(avp)
            padding = avp.get_padding_length()
            if padding:
                real_length += avp.get_padding_length()

        if real_length != self.get_length():
            self.header.length = convert_to_3_bytes(real_length)


class AddressType(OctetStringType):
    @abc.abstractmethod
    def __init__(self, data, vendor_id=None):
        AddressType.parser_data(self, data)
        OctetStringType.__init__(self, data, vendor_id)


    def parser_data(self, data):
        if isinstance(data, bytes):
            if data[:2] == HOST_IP_ADDRESS_FAMILY_CODE_IPV4:
                try:
                    ipaddress.IPv4Address(data[2:])
                except ipaddress.AddressValueError:
                    raise DataTypeError("Stream of bytes does not "\
                                        "correspond to a valid IPv4 address "\
                                        "format")

            elif data[:2] == HOST_IP_ADDRESS_FAMILY_CODE_IPV6:
                try:
                    ipaddress.IPv6Address(data[2:])
                except ipaddress.AddressValueError:
                    raise DataTypeError("Stream of bytes does not "\
                                        "correspond to a valid IPv6 address "\
                                        "format")

        else:
            ip_address = ipaddress.ip_address(data)
    
            if isinstance(ip_address, ipaddress.IPv4Address):
                ip_address_family_code = HOST_IP_ADDRESS_FAMILY_CODE_IPV4

            elif isinstance(ip_address, ipaddress.IPv6Address):
                ip_address_family_code = HOST_IP_ADDRESS_FAMILY_CODE_IPV6

            else:
                raise DataTypeError("AddressType MUST have data argument "\
                                    "of 'str' with IP address format value")

            data = ip_address_family_code + ip_address.packed
        
        self._data = data                


    def is_ipv4(self):
        ip_address = ipaddress.ip_address(self.data[2:])

        if isinstance(ip_address, ipaddress.IPv4Address):
            return True
        return False


    def is_ipv6(self):
        ip_address = ipaddress.ip_address(self.data[2:])

        if isinstance(ip_address, ipaddress.IPv6Address):
            return True
        return False

    
    def get_ip_address(self):
        if self.is_ipv4():
            return str(ipaddress.IPv4Address(self.data[2:]))

        if self.is_ipv6():
            return str(ipaddress.IPv6Address(self.data[2:]))


class TimeType(OctetStringType):
    @abc.abstractmethod
    def __init__(self, data, vendor_id=None):
        if isinstance(data, bytes):
            if len(data) != 4:
                raise DataTypeError("Invalid data format for TimeType. It "\
                                    "MUST be 4 bytes long")

        elif isinstance(data, datetime.datetime):
            ref = datetime.datetime(1900, 1, 1, 0, 0, 0)
            diff = data - ref

            timestamp = diff.days*24*60*60 + diff.seconds
            data = convert_to_4_bytes(timestamp)
        
        elif not isinstance(data, datetime.datetime) or not isinstance(data, bytes):
            raise DataTypeError("Invalid data format for TimeType. It MUST "\
                                "be a datetime object")

        OctetStringType.__init__(self, data, vendor_id)    


class UTF8StringType(OctetStringType):
    @abc.abstractmethod
    def __init__(self, data, vendor_id=None):
        OctetStringType.__init__(self, data, vendor_id)


class DiameterIdentityType(OctetStringType):
    @abc.abstractmethod
    def __init__(self, data, vendor_id=None):
        OctetStringType.__init__(self, data, vendor_id)


class DiameterURIType(OctetStringType):
    @abc.abstractmethod
    def __init__(self, data, vendor_id=None):
        self.parser_data(data)
        OctetStringType.__init__(self, data, vendor_id)


    def parser_data(self, data):
        if isinstance(data, bytes):
            data = data.decode("utf-8")

        elif isinstance(data, str):
            data = data

        else:
            raise DataTypeError("invalid data format. Data MUST be of "\
                                "'str' or 'bytes' and follow the URI syntax")

        pattern = r"aaa[s]{0,1}://(?!\d\.)[a-zA-Z1-9_\-].{1,62}"\
                  r"[a-zA-Z1-9_\-](\:\b([1-9]|[1-9][0-9]|[1-9]"\
                  r"[0-9][0-9]|[1-9][0-9][0-9][0-9]|[1-3][0-9]"\
                  r"[0-9][0-9][0-9]|4[0-8][0-9][0-9][0-9]|490[0-9]"\
                  r"[0-9]|491[0-4][0-9]|49150|49151)\b){0,1}(;transport"\
                  r"=(tcp|udp|sctp)){0,1}(;protocol=(diameter|radius)){0,1}"

        if not re.fullmatch(pattern, data):
            raise DataTypeError("invalid data format. It does not comply "\
                                "to the DiameterURI syntax")

        self._data = data.encode("utf-8")


class EnumeratedType(Integer32Type):
    @abc.abstractmethod
    def __init__(self, data, vendor_id=None):
        avp_class_name = self.__class__.__name__

        if data not in self.values:
            raise AVPAttributeValueError(f"invalid input argument for "\
                                         f"{avp_class_name}. Refer to the "\
                                         f"`values` class attribute for options")

        Integer32Type.__init__(self, data, vendor_id)



# -*- coding: utf-8 -*-
"""
    bromelia.constants.ietf
    ~~~~~~~~~~~~~~~~~~~~~~~

    This module contains constants defined in IETF specs.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""



# -*- coding: utf-8 -*-
"""
    bromelia.lib.etsi_3gpp_swm
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    A Python package with APIs for handling 3GPP SWm Application Id in the
    bromelia Python Library context.

    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from .avps import *

from .definitions import *

from .messages import AbortSessionAnswer as ASA
from .messages import AbortSessionRequest as ASR

from .messages import DiameterEapAnswer as DEA
from .messages import DiameterEapRequest as DER


# -*- coding: utf-8 -*-
"""
    bromelia.process
    ~~~~~~~~~~~~~~~~

    This module contains all implementations for handling the Diameter 
    messages processing step. It contains classes for creating and 
    parsing Diameter Requests and Answers.

    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""
import logging

from .avps import AuthRequestTypeAVP
from .constants import *
from .exceptions import ProcessRequestException

process_message_logging = logging.getLogger("ProcessDiameterMessage")


def process_request(association, message):
    list_of_avps_by_code = [avp.code for avp in message.avps]

    connection = association.connection

    local_node_host_name = connection.local_node.host_name.encode("utf-8")
    local_node_realm = connection.local_node.realm.encode("utf-8")

    if DESTINATION_HOST_AVP_CODE in list_of_avps_by_code:
        if not list(filter(lambda avp: avp.data == local_node_host_name, message.avps)):
            logging.debug(f"[{message.header.hop_by_hop.hex()}] Diameter "\
                          f"Request has Destination-Host AVP, however it was "\
                          f"addressed to another node.")
            
            raise ProcessRequestException("Request does not comply with "\
                                          "local consumption rules.")

        logging.debug(f"[{message.header.hop_by_hop.hex()}] Diameter Request "\
                      f"has Destination-Host AVP and contains the hostname of "\
                      f"local node.")

    elif (DESTINATION_HOST_AVP_CODE not in list_of_avps_by_code and 
          DESTINATION_REALM_AVP_CODE in list_of_avps_by_code):

        if not list(filter(lambda avp: avp.data == local_node_realm, message.avps)):
            logging.debug(f"[{message.header.hop_by_hop.hex()}] Diameter "\
                          f"Request does not include Destination-Host AVP, "\
                          f"but it does include an invalid Destination-Realm "\
                          f"AVP which was addressed to another realm.")

            raise ProcessRequestException("Request does not comply with "\
                                          "local consumption rules.")

        logging.debug(f"[{message.header.hop_by_hop.hex()}] Diameter Request "\
                      f"does not include Destination-Host AVP, but it does "\
                      f"include a valid Destination-Realm AVP which contains "\
                      f"realm of local node.")
        
    elif (DESTINATION_HOST_AVP_CODE not in list_of_avps_by_code and 
          DESTINATION_REALM_AVP_CODE not in list_of_avps_by_code):

        logging.debug(f"[{message.header.hop_by_hop.hex()}] Diameter Request "\
                      f"does not include neither Destination-Host AVP nor "\
                      f"Destination-Realm AVP.")

    else:
        raise ProcessRequestException("Request does not comply with local "\
                                      "consumption rules.")

    association.num_requests += 1
    logging.debug(f"[{message.header.hop_by_hop.hex()}] Processed Diameter Request.")


def process_answer(association, message):
    end_to_end_key = message.header.end_to_end.hex()
    hop_by_hop_key = message.header.hop_by_hop.hex()

    if end_to_end_key in association.end_to_end_identifiers:
        if hop_by_hop_key in association.pending_requests:
            if message.header.end_to_end == association.pending_requests[hop_by_hop_key].header.end_to_end:
                association.pending_requests.pop(hop_by_hop_key)
                association.num_answers += 1
    
    logging.debug(f"[{message.header.hop_by_hop.hex()}] Processed Diameter Answer.")


class ProcessDiameterMessage:
    @staticmethod
    def process_answer_from_existing_pending_request(association, message):
        process_message_logging.debug("Processing Diameter Answer.")

        end_to_end_key = message.header.end_to_end.hex()
        hop_by_hop_key = message.header.hop_by_hop.hex()

        if end_to_end_key in association.end_to_end_identifiers:
            if hop_by_hop_key in association.pending_requests:
                if message.header.end_to_end == association.pending_requests[hop_by_hop_key].header.end_to_end:
                    association.pending_requests.pop(hop_by_hop_key)

    
    @staticmethod
    def is_valid_origin_host_avp(avp, connection):
        if avp.code == ORIGIN_HOST_AVP_CODE:
            
            process_message_logging.debug(f"Origin-Host AVP validation.")

            checklist_mandatory_info = 0
        
            process_message_logging.debug(f"flags: {avp.get_flags()}.")
            if avp.flags == FLAG_NOT_VENDOR_SPECIFIC_AND_MANDATORY_AND_NOT_PROTECTED:
                checklist_mandatory_info += 1
            
            process_message_logging.debug(f"length: {avp.get_length()}.")
            if avp.get_length() == AVP_HEADER_LENGTH + len(avp.data):
                checklist_mandatory_info += 1

            data = avp.data.decode("utf-8")
            process_message_logging.debug(f"data: {data}.")
            if data == connection.peer_node.host_name:
                checklist_mandatory_info += 1


            if checklist_mandatory_info == 3:
                process_message_logging.debug(f"Result: PASS.")
                return True
            process_message_logging.debug(f"Result: FAIL.")
            return False


    @staticmethod
    def is_valid_origin_realm_avp(avp, connection):
        if avp.code == ORIGIN_REALM_AVP_CODE:
            process_message_logging.debug(f"Origin-Realm AVP validation.")

            checklist_mandatory_info = 0
        
            process_message_logging.debug(f"flags: {avp.get_flags()}.")
            if avp.flags == FLAG_NOT_VENDOR_SPECIFIC_AND_MANDATORY_AND_NOT_PROTECTED:
                checklist_mandatory_info += 1
            
            process_message_logging.debug(f"length: {avp.get_length()}.")
            if avp.get_length() == AVP_HEADER_LENGTH + len(avp.data):
                checklist_mandatory_info += 1

            data = avp.data.decode("utf-8")
            process_message_logging.debug(f"data: {data}.")
            if data == connection.peer_node.realm:
                checklist_mandatory_info += 1


            if checklist_mandatory_info == 3:
                process_message_logging.debug(f"Result: PASS.")
                return True
            process_message_logging.debug(f"Result: FAIL.")
            return False


    @staticmethod
    def is_valid_origin_state_id_avp(avp, connection):
        if (avp.code == ORIGIN_STATE_ID_AVP_CODE):
            return True
        return False


    @staticmethod
    def is_valid_result_code_avp(avp):
        if avp.code == RESULT_CODE_AVP_CODE:
            
            process_message_logging.debug(f"Result-Code AVP validation.")

            checklist_mandatory_info = 0
        
            process_message_logging.debug(f"flags: {avp.get_flags()}.")
            if avp.flags == FLAG_NOT_VENDOR_SPECIFIC_AND_MANDATORY_AND_NOT_PROTECTED:
                checklist_mandatory_info += 1
            
            process_message_logging.debug(f"length: {avp.get_length()}.")
            if avp.get_length() == AVP_LENGTH_UNSIGNED32:
                checklist_mandatory_info += 1

            process_message_logging.debug(f"data: {avp.data.hex()}.")


            if checklist_mandatory_info == 2:
                process_message_logging.debug(f"Result: PASS.")
                return True
            process_message_logging.debug(f"Result: FAIL.")
            return False


    @staticmethod
    def is_valid_disconnect_cause_avp(avp):
        if (avp.code == DISCONNECT_CAUSE_AVP_CODE) and (avp.data == DISCONNECT_CAUSE_REBOOTING):
            return True
        return False


    @staticmethod
    def is_valid_host_ip_address_avp(avp, connection):
        if (avp.code == HOST_IP_ADDRESS_AVP_CODE):
            host_ip_address = "{}.{}.{}.{}".format(int(avp.data[2]),int(avp.data[3]),int(avp.data[4]),int(avp.data[5]))
            return True
            # if connection.peer_node.ip_address == host_ip_address:
            #     return True
            # process_message_logging.debug("Host IP Address from Peer Node not valid!")
            # return False


    @staticmethod
    def is_valid_product_name_avp(avp, connection):
        if (avp.code == PRODUCT_NAME_AVP_CODE):
            return True
        return False


    @staticmethod
    def is_valid_vendor_id_avp(avp, connection):
        if (avp.code == VENDOR_ID_AVP_CODE):
            return True
        return False


    @staticmethod
    def is_valid_session_id_avp(avp):
        if avp.code == SESSION_ID_AVP_CODE:
            process_message_logging.debug(f"Session-Id AVP validation.")

            checklist_mandatory_info = 0
        
            process_message_logging.debug(f"flags: {avp.get_flags()}.")
            if avp.flags == FLAG_NOT_VENDOR_SPECIFIC_AND_MANDATORY_AND_NOT_PROTECTED:
                checklist_mandatory_info += 1
            
            process_message_logging.debug(f"length: {avp.get_length()}.")
            if avp.get_length() == (AVP_HEADER_LENGTH + len(avp.data)):
                checklist_mandatory_info += 1

            process_message_logging.debug(f"data: {avp.data.hex()}.")


            if checklist_mandatory_info == 2:
                process_message_logging.debug(f"Result: PASS.")
                return True
            process_message_logging.debug(f"Result: FAIL.")
            return False


    @staticmethod
    def is_valid_auth_application_id_avp(avp):
        if avp.code == AUTH_APPLICATION_ID_AVP_CODE:
            process_message_logging.debug(f"Auth-Application-Id AVP validation.")

            checklist_mandatory_info = 0
        
            process_message_logging.debug(f"flags: {avp.get_flags()}.")
            if avp.flags == FLAG_NOT_VENDOR_SPECIFIC_AND_MANDATORY_AND_NOT_PROTECTED:
                checklist_mandatory_info += 1

            process_message_logging.debug(f"length: {avp.get_length()}.")
            if avp.get_length() == AVP_LENGTH_UNSIGNED32:
                checklist_mandatory_info += 1
        

            if checklist_mandatory_info == 2: # == 3
                process_message_logging.debug(f"Result: PASS.")
                return True
            process_message_logging.debug(f"Result: FAIL.")
            return False


    @staticmethod
    def is_valid_auth_request_type_avp(avp):
        if avp.code == AUTH_REQUEST_TYPE_AVP_CODE:
            process_message_logging.debug(f"Auth-Request-Type AVP validation.")

            checklist_mandatory_info = 0
        
            process_message_logging.debug(f"flags: {avp.get_flags()}.")
            if avp.flags == FLAG_NOT_VENDOR_SPECIFIC_AND_MANDATORY_AND_NOT_PROTECTED:
                checklist_mandatory_info += 1
            
            process_message_logging.debug(f"length: {avp.get_length()}.")
            if avp.get_length() == AVP_LENGTH_INTEGER32:
                checklist_mandatory_info += 1

            process_message_logging.debug(f"data: {avp.data.hex()}.")
            if avp.data in AuthRequestTypeAVP.values:
                checklist_mandatory_info += 1


            if checklist_mandatory_info == 3:
                process_message_logging.debug(f"Result: PASS.")
                return True
            process_message_logging.debug(f"Result: FAIL.")
            return False


class ProcessDeviceWatchdog():
    def __init__(self, association, message):
        self.association = association
        self.connection = association.connection
        self.message = message

        self.checklist_mandatory_avps = 0
        self.checklist_optional_avps = 0
        self.checklist_error_avps = 0
        self.is_valid = False

        if message.header.flags == FLAG_REQUEST:
            self.process_request()
        elif message.header.flags == FLAG_RESPONSE:
            self.process_answer()
        else:
            """What should we do if find an error?"""
            pass


    def process_request(self):
        for avp in self.message.avps:
            if ProcessDiameterMessage.is_valid_origin_host_avp(avp, self.connection):
                self.checklist_mandatory_avps += 1

            elif ProcessDiameterMessage.is_valid_origin_realm_avp(avp, self.connection):
                self.checklist_mandatory_avps += 1

            elif ProcessDiameterMessage.is_valid_origin_state_id_avp(avp, self.connection):
                self.checklist_optional_avps += 1


        if (self.checklist_mandatory_avps == 2) and (self.checklist_optional_avps == 0 or self.checklist_optional_avps == 1):
            self.is_valid = True
        else:
            self.is_valid = False


    def process_answer(self):
        ProcessDiameterMessage.process_answer_from_existing_pending_request(self.association, self.message)

        for avp in self.message.avps:
            if ProcessDiameterMessage.is_valid_result_code_avp(avp):
                self.checklist_mandatory_avps += 1

            if ProcessDiameterMessage.is_valid_origin_host_avp(avp, self.connection):
                self.checklist_mandatory_avps += 1

            elif ProcessDiameterMessage.is_valid_origin_realm_avp(avp, self.connection):
                self.checklist_mandatory_avps += 1

            elif ProcessDiameterMessage.is_valid_origin_state_id_avp(avp, self.connection):
                self.checklist_optional_avps += 1


        if (self.checklist_mandatory_avps == 3) and (self.checklist_optional_avps == 0 or self.checklist_optional_avps == 1):
            self.is_valid = True
        else:
            self.is_valid = False


class ProcessDisconnectPeer():
    def __init__(self, association, message):
        self.association = association
        self.connection = association.connection
        self.message = message

        self.checklist_mandatory_avps = 0
        self.checklist_optional_avps = 0
        self.checklist_error_avps = 0
        self.is_valid = False

        if message.header.flags == FLAG_REQUEST:
            self.process_request()
        elif message.header.flags == FLAG_RESPONSE:
            self.process_answer()
        else:
            """What should we do if find an error?"""
            pass


    def process_request(self):
        for avp in self.message.avps:
            if ProcessDiameterMessage.is_valid_origin_host_avp(avp, self.connection):
                self.checklist_mandatory_avps += 1

            elif ProcessDiameterMessage.is_valid_origin_realm_avp(avp, self.connection):
                self.checklist_mandatory_avps += 1

            elif ProcessDiameterMessage.is_valid_disconnect_cause_avp(avp):
                self.checklist_mandatory_avps += 1


        if (self.checklist_mandatory_avps == 3):
            self.is_valid = True
        else:
            self.is_valid = False


    def process_answer(self):
        ProcessDiameterMessage.process_answer_from_existing_pending_request(self.association, self.message)

        for avp in self.message.avps:
            if ProcessDiameterMessage.is_valid_result_code_avp(avp):
                self.checklist_mandatory_avps += 1

            if ProcessDiameterMessage.is_valid_origin_host_avp(avp, self.connection):
                self.checklist_mandatory_avps += 1

            elif ProcessDiameterMessage.is_valid_origin_realm_avp(avp, self.connection):
                self.checklist_mandatory_avps += 1


        if (self.checklist_mandatory_avps == 3) and (self.checklist_error_avps >= 0 and self.checklist_error_avps <= 2):
            self.is_valid = True
        else:
            self.is_valid = False


class ProcessCapabilityExchange():
    def __init__(self, association, message):
        self.association = association
        self.connection = association.connection
        self.message = message

        self.checklist_mandatory_avps = 0
        self.checklist_optional_avps = 0
        self.checklist_error_avps = 0
        self.is_valid = False

        if message.header.flags == FLAG_REQUEST:
            self.process_request()
        elif message.header.flags == FLAG_RESPONSE:
            self.process_answer()
        else:
            """What should we do if find an error?"""
            pass


    def process_request(self):
        for avp in self.message.avps:
            if ProcessDiameterMessage.is_valid_origin_host_avp(avp, self.connection):
                self.checklist_mandatory_avps += 1

            elif ProcessDiameterMessage.is_valid_origin_realm_avp(avp, self.connection):
                self.checklist_mandatory_avps += 1

            elif ProcessDiameterMessage.is_valid_host_ip_address_avp(avp, self.connection):
                self.checklist_mandatory_avps += 1

            elif ProcessDiameterMessage.is_valid_vendor_id_avp(avp, self.connection):
                self.checklist_mandatory_avps += 1

            elif ProcessDiameterMessage.is_valid_product_name_avp(avp, self.connection):
                self.checklist_mandatory_avps += 1

            elif ProcessDiameterMessage.is_valid_origin_state_id_avp(avp, self.connection):
                self.checklist_optional_avps += 1


        if (self.checklist_mandatory_avps == 5) and (self.checklist_optional_avps >= 0 and self.checklist_optional_avps <= 7):
            self.is_valid = True
        else:
            self.is_valid = False


    def process_answer(self):
        ProcessDiameterMessage.process_answer_from_existing_pending_request(self.association, self.message)
        for avp in self.message.avps:
            if ProcessDiameterMessage.is_valid_result_code_avp(avp):
                self.checklist_mandatory_avps += 1

            if ProcessDiameterMessage.is_valid_origin_host_avp(avp, self.connection):
                self.checklist_mandatory_avps += 1

            elif ProcessDiameterMessage.is_valid_origin_realm_avp(avp, self.connection):
                self.checklist_mandatory_avps += 1

            elif ProcessDiameterMessage.is_valid_host_ip_address_avp(avp, self.connection):
                self.checklist_mandatory_avps += 1

            elif ProcessDiameterMessage.is_valid_vendor_id_avp(avp, self.connection):
                self.checklist_mandatory_avps += 1

            elif ProcessDiameterMessage.is_valid_product_name_avp(avp, self.connection):
                self.checklist_mandatory_avps += 1

            elif ProcessDiameterMessage.is_valid_origin_state_id_avp(avp, self.connection):
                self.checklist_optional_avps += 1


        if (self.checklist_mandatory_avps == 6) and (self.checklist_optional_avps >= 0 or self.checklist_optional_avps <= 7):
            self.is_valid = True
        else:
            self.is_valid = False


def process_device_watchdog(association, message):
    return ProcessDeviceWatchdog(association, message)


def process_disconnect_peer(association, message):
    return ProcessDisconnectPeer(association, message)


def process_capability_exchange(association, message):
    return ProcessCapabilityExchange(association, message)




# -*- coding: utf-8 -*-
"""
    bromelia.etsi_3gpp_s13.avps
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains the Diameter protocol AVP library that are used 
    to create Diameter messages for 3GPP S13/S13' Application Id.
    
    :copyright: (c) 2021-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from .._internal_utils import show_warn

show_warn("avps", "etsi_3gpp_s13")

from ..avps.etsi_3gpp.ts_129_272 import TerminalInformationAVP
from ..avps.etsi_3gpp.ts_129_272 import EquipmentStatusAVP

from ..avps.ietf.rfc6733 import AuthApplicationIdAVP
from ..avps.ietf.rfc6733 import AuthSessionStateAVP
from ..avps.ietf.rfc6733 import DestinationHostAVP
from ..avps.ietf.rfc6733 import DestinationRealmAVP
from ..avps.ietf.rfc6733 import ExperimentalResultAVP
from ..avps.ietf.rfc6733 import FailedAvpAVP
from ..avps.ietf.rfc6733 import OriginHostAVP
from ..avps.ietf.rfc6733 import OriginRealmAVP
from ..avps.ietf.rfc6733 import ProxyInfoAVP
from ..avps.ietf.rfc6733 import ResultCodeAVP
from ..avps.ietf.rfc6733 import RouteRecordAVP
from ..avps.ietf.rfc6733 import SessionIdAVP
from ..avps.ietf.rfc6733 import UserNameAVP
from ..avps.ietf.rfc6733 import VendorIdAVP
from ..avps.ietf.rfc6733 import VendorSpecificApplicationIdAVP


# -*- coding: utf-8 -*-
"""
    bromelia.constants.flags
    ~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains list of Diameter Flags defined in IETF RFC 6733.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from .._internal_utils import convert_to_1_byte


#: List of COMMAND FLAGS IN DIAMETER HEADER.
FLAG_RESPONSE = convert_to_1_byte(0x00)
FLAG_RESPONSE_AND_RETRANSMITTED = convert_to_1_byte(0x10)
FLAG_RESPONSE_AND_ERROR = convert_to_1_byte(0x20)
FLAG_RESPONSE_AND_ERROR_AND_RETRANSMITTED = convert_to_1_byte(0x30)
FLAG_RESPONSE_AND_PROXYABLE = convert_to_1_byte(0x40)
FLAG_RESPONSE_AND_PROXYABLE_AND_RETRANSMITTED = convert_to_1_byte(0x50)
FLAG_RESPONSE_AND_PROXYABLE_AND_ERROR = convert_to_1_byte(0x60)
FLAG_RESPONSE_AND_PROXYABLE_AND_ERROR_AND_RETRANSMITTED = convert_to_1_byte(0x70)

FLAG_REQUEST = convert_to_1_byte(0x80)
FLAG_REQUEST_AND_RETRANSMITTED = convert_to_1_byte(0x90)
FLAG_REQUEST_AND_ERROR = convert_to_1_byte(0xa0)
FLAG_REQUEST_AND_ERROR_AND_RETRANSMITTED = convert_to_1_byte(0xb0)
FLAG_REQUEST_AND_PROXYABLE = convert_to_1_byte(0xc0)
FLAG_REQUEST_AND_PROXYABLE_AND_RETRANSMITTED = convert_to_1_byte(0xd0)
FLAG_REQUEST_AND_PROXYABLE_AND_ERROR = convert_to_1_byte(0xe0)
FLAG_REQUEST_AND_PROXYABLE_AND_ERROR_AND_RETRANSMITTED = convert_to_1_byte(0xf0)

#: List of COMMAND FLAGS IN DIAMETER AVP HEADER.
FLAG_NOT_VENDOR_SPECIFIC_AND_NOT_MANDATORY_AND_NOT_PROTECTED = convert_to_1_byte(0x00)
FLAG_NOT_VENDOR_SPECIFIC_AND_NOT_MANDATORY_AND_PROTECTED = convert_to_1_byte(0x20)
FLAG_NOT_VENDOR_SPECIFIC_AND_MANDATORY_AND_NOT_PROTECTED = convert_to_1_byte(0x40)
FLAG_NOT_VENDOR_SPECIFIC_MANDATORY_AND_PROTECTED = convert_to_1_byte(0x60)

FLAG_VENDOR_SPECIFIC_AND_NOT_MANDATORY_AND_NOT_PROTECTED = convert_to_1_byte(0x80)
FLAG_VENDOR_SPECIFIC_AND_NOT_MANDATORY_AND_PROTECTED = convert_to_1_byte(0xa0)
FLAG_VENDOR_SPECIFIC_AND_MANDATORY_AND_NOT_PROTECTED = convert_to_1_byte(0xc0)
FLAG_VENDOR_SPECIFIC_MANDATORY_AND_PROTECTED = convert_to_1_byte(0xe0)



# -*- coding: utf-8 -*-
"""
    bromelia
    ~~~~~~~~

    A Python microframework for building Diameter protocol applications.

    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from .__version__ import __title__
from .__version__ import __description__
from .__version__ import __url__
from .__version__ import __version__
from .__version__ import __author__
from .__version__ import __author_email__
from .__version__ import __license__
from .__version__ import __copyright__

from .base import *
from .bromelia import Bromelia
from .setup import Diameter


# -*- coding: utf-8 -*-
"""
    bromelia.lib.etsi_3gpp_s6a.avps
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains the Diameter protocol AVP library that are used 
    to create Diameter messages for 3GPP S6a/S6d Application Id.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from ...avps.etsi_3gpp.ts_129_212 import RatTypeAVP

from ...avps.etsi_3gpp.ts_129_229 import SupportedFeaturesAVP
from ...avps.etsi_3gpp.ts_129_229 import VisitedNetworkIdentifierAVP

from ...avps.etsi_3gpp.ts_129_272 import ContextIdentifierAVP
from ...avps.etsi_3gpp.ts_129_272 import ServiceSelectionAVP
from ...avps.etsi_3gpp.ts_129_272 import UeUsageTypeAVP
from ...avps.etsi_3gpp.ts_129_272 import SubscriptionDataAVP
from ...avps.etsi_3gpp.ts_129_272 import RequestedEutranAuthenticationInfoAVP
from ...avps.etsi_3gpp.ts_129_272 import TerminalInformationAVP
from ...avps.etsi_3gpp.ts_129_272 import UlrFlagsAVP
from ...avps.etsi_3gpp.ts_129_272 import UlaFlagsAVP
from ...avps.etsi_3gpp.ts_129_272 import AirFlagsAVP
from ...avps.etsi_3gpp.ts_129_272 import NorFlagsAVP
from ...avps.etsi_3gpp.ts_129_272 import PurFlagsAVP
from ...avps.etsi_3gpp.ts_129_272 import PuaFlagsAVP
from ...avps.etsi_3gpp.ts_129_272 import AlertReasonAVP
from ...avps.etsi_3gpp.ts_129_272 import ErrorDiagnosticAVP
from ...avps.etsi_3gpp.ts_129_272 import VisitedPlmnIdAVP
from ...avps.etsi_3gpp.ts_129_272 import AuthenticationInfoAVP
from ...avps.etsi_3gpp.ts_129_272 import HomogeneousSupportOfImsVoiceOverPsSessionsAVP
from ...avps.etsi_3gpp.ts_129_272 import UeSrvccCapabilityAVP
from ...avps.etsi_3gpp.ts_129_272 import SupportedServicesAVP
from ...avps.etsi_3gpp.ts_129_272 import CancellationTypeAVP
from ...avps.etsi_3gpp.ts_129_272 import ClrFlagsAVP

from ...avps.ietf.rfc5447 import Mip6AgentInfoAVP

from ...avps.ietf.rfc6733 import AuthApplicationIdAVP
from ...avps.ietf.rfc6733 import AuthSessionStateAVP
from ...avps.ietf.rfc6733 import DestinationHostAVP
from ...avps.ietf.rfc6733 import DestinationRealmAVP
from ...avps.ietf.rfc6733 import ExperimentalResultAVP
from ...avps.ietf.rfc6733 import FailedAvpAVP
from ...avps.ietf.rfc6733 import OriginHostAVP
from ...avps.ietf.rfc6733 import OriginRealmAVP
from ...avps.ietf.rfc6733 import ProxyInfoAVP
from ...avps.ietf.rfc6733 import ResultCodeAVP
from ...avps.ietf.rfc6733 import RouteRecordAVP
from ...avps.ietf.rfc6733 import SessionIdAVP
from ...avps.ietf.rfc6733 import UserNameAVP
from ...avps.ietf.rfc6733 import VendorIdAVP
from ...avps.ietf.rfc6733 import VendorSpecificApplicationIdAVP


# -*- coding: utf-8 -*-
"""
    bromelia.lib.ietf_rfc6733
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    A Python package with APIs for handling Diameter protocol base messages
    defined in IETF RFC 6733.

    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from .avps import *

from .messages import AbortSessionAnswer as ASA
from .messages import AbortSessionRequest as ASR

from .messages import CapabilitiesExchangeAnswer as CEA
from .messages import CapabilitiesExchangeRequest as CER

from .messages import DeviceWatchdogAnswer as DWA
from .messages import DeviceWatchdogRequest as DWR

from .messages import DisconnectPeerAnswer as DPA
from .messages import DisconnectPeerRequest as DPR

from .messages import ReAuthAnswer as RAA
from .messages import ReAuthRequest as RAR

from .messages import SessionTerminationAnswer as STA
from .messages import SessionTerminationRequest as STR


# -*- coding: utf-8 -*-
"""
    bromelia.transport
    ~~~~~~~~~~~~~~~~~~
    
    This module defines the TCP transport layer connections that are used
    by the Diameter application protocol underlying.

    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

import copy
import logging
import random
import selectors
import socket
import threading
from .config import TRACKING_SOCKET_EVENTS_TIMEOUT

tcp_connection = logging.getLogger("TcpConnection")
tcp_client = logging.getLogger("TcpClient")
tcp_server = logging.getLogger("TcpServer")


class TcpConnection():
    def __init__(self, ip_address, port):
        self._recv_buffer = b""
        self._send_buffer = b""
        self.send_data_stream_queued = False
        self.data_stream = b""
        self._recv_data_stream = b""

        self._recv_data_available = threading.Event()
        self.write_mode_on = threading.Event()
        self.read_mode_on = threading.Event()

        self.lock = threading.Lock()

        self.recv_data_consumed = False
        
        self.ip_address = ip_address
        self.port = port

        self.sock = None
        self.is_connected = False
        self.sock_id = "".join(random.choice('0123456789ABCDEF') for i in range(16))

        tcp_connection.debug(f"Creating Socket with ID {self.sock_id}")

        self._stop_threads = False

        self.selector = selectors.DefaultSelector()
        self.tracking_events_count = 0
        self.connection_attempts = 3

        self.events_mask = selectors.EVENT_READ
        

    def is_write_mode(self):
        if self.events_mask & selectors.EVENT_WRITE:
            return True
        return False

    
    def is_read_mode(self):
        if self.events_mask & selectors.EVENT_READ:
            return True
        return False

    def is_read_write_mode(self):
        if self.events_mask & (selectors.EVENT_READ | selectors.EVENT_WRITE):
            return True
        return False
   

    def close(self):
        if not self.is_connected:
            raise ConnectionError("There is no transport connection up for "\
                                  "this PeerNode")

        self.is_connected = False
        try:
            self.selector.unregister(self.sock)
            tcp_connection.debug(f"[Socket-{self.sock_id}] De-registering "\
                                 f"Socket from Selector address: "\
                                 f"{self.selector.get_map()}")
    
            self.sock.close()
            tcp_connection.debug(f"[Socket-{self.sock_id}] Shutting "\
                                 f"down Socket")

        except KeyError as e:
            tcp_connection.debug(f"[Socket-{self.sock_id}] There is no "\
                                 f"such Selector registered")

        self._stop_threads = True


    def run(self):
        if not self.is_connected:
            raise ConnectionError(f"[Socket-{self.sock_id}] There is no "\
                                  f"transport connection up for this Peer")

        threading.Thread(name="transport_layer_bootstrapper", 
                         target=self._run).start()


    def _run(self):
        while self.is_connected and not self._stop_threads:
            self.events = self.selector.select(timeout=TRACKING_SOCKET_EVENTS_TIMEOUT)
            self.tracking_events_count += TRACKING_SOCKET_EVENTS_TIMEOUT

            for key, mask in self.events:
                if key.data is not None:
                    self.data_stream += key.data

                if mask & selectors.EVENT_WRITE:
                    tcp_connection.debug(f"Selector notified EVENT_WRITE")
                    self.write()

                if mask & selectors.EVENT_READ:
                    tcp_connection.debug(f"Selector notified EVENT_READ")
                    self.read()


    def _set_selector_events_mask(self, mode, msg=None):
        self.lock.acquire()
        if mode == "r":
            tcp_connection.debug(f"[Socket-{self.sock_id}] Updating "\
                                 f"selector events mask [READ]")

            self.events_mask = selectors.EVENT_READ
            self.selector.modify(self.sock, self.events_mask)
            self.write_mode_on.clear()
            self.read_mode_on.set()
            
        elif mode == "w":
            tcp_connection.debug(f"[Socket-{self.sock_id}] Updating "\
                                 f"selector events mask [WRITE]")

            self.events_mask = selectors.EVENT_WRITE
            self.selector.modify(self.sock, self.events_mask, data=msg)
            self.write_mode_on.set()
            self.read_mode_on.clear()


        elif mode == "rw":
            tcp_connection.debug(f"[Socket-{self.sock_id}] Updating "\
                                 f"selector events mask [READ/WRITE]")

            self.events_mask = selectors.EVENT_READ | selectors.EVENT_WRITE
            self.selector.modify(self.sock, self.events_mask, data=msg)
            self.write_mode_on.set()
            self.read_mode_on.set()

        else:
            tcp_connection.debug(f"[Socket-{self.sock_id}] Updating "\
                                 f"selector events mask: Invalid entry")
        self.lock.release()


    def _write(self):
        if self._send_buffer:
            try:
                sent = self.sock.send(self._send_buffer)
                tcp_connection.debug(f"[Socket-{self.sock_id}] Just sent "\
                                     f"{sent} bytes in _send_buffer")
            
            except BlockingIOError:
                tcp_connection.exception(f"[Socket-{self.sock_id}] An error "\
                                         f"has occurred")

                self._stop_threads = True
                
            else:
                self._send_buffer = self._send_buffer[sent:]

            tcp_connection.debug(f"[Socket-{self.sock_id}] Stream data "\
                                 f"has been sent")


    def write(self):
        if not self.send_data_stream_queued and self.data_stream:
            self._send_buffer += self.data_stream
            self.data_stream = b""
            self.send_data_stream_queued = True
            tcp_connection.debug(f"[Socket-{self.sock_id}] Stream data has "\
                                 f"been queued into _send_buffer: "\
                                 f"{self._send_buffer.hex()}")

        self._write()

        if self.send_data_stream_queued and not self._send_buffer:
            self._set_selector_events_mask("r")
            self.send_data_stream_queued = False
            tcp_connection.debug(f"[Socket-{self.sock_id}] There is no "\
                                 f"data to be sent for a while")


    def _read(self):
        try:
            data = self.sock.recv(4096*64)
            tcp_connection.debug(f"[Socket-{self.sock_id}] Data received: "\
                                 f"{data.hex()}")

        except:
            tcp_connection.exception(f"[Socket-{self.sock_id}] An Exception "\
                                     f"has been raised")

            self.error_has_raised = True
            self._stop_threads = True

        else:
            if data:
                self._recv_buffer += data
                tcp_connection.debug(f"[Socket-{self.sock_id}] _recv_buffer: "\
                                     f"{self._recv_buffer.hex()}")
            else:
                tcp_connection.debug(f"[Socket-{self.sock_id}] Peer closed "\
                                     f"connection")
                self._stop_threads = True


    def read(self):
        self._read()

        if self._recv_buffer:
            self._recv_data_stream += copy.copy(self._recv_buffer)
            self._recv_data_available.set()
            self._recv_buffer = b""

        tcp_connection.debug(f"[Socket-{self.sock_id}] _recv_buffer has "\
                             f"been cleaned up")

        self._set_selector_events_mask("r")


    def test_connection(self):
        while True:
            try:
                self.sock.send(b"")
                return True

            except OSError as e:
                if e.args[0] == 10057:
                    self.connection_attempts -= self.connection_attempts
                    return False


class TcpClient(TcpConnection):
    def __init__(self, ip_address, port):
        super().__init__(ip_address, port)


    def start(self):
        try:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            tcp_client.debug(f"[Socket-{self.sock_id}] Client-side Socket: "\
                             f"{self.sock}")

            self.sock.setblocking(False)
            tcp_client.debug(f"[Socket-{self.sock_id}] Setting as "\
                             f"Non-Blocking")

            self.sock.connect_ex((self.ip_address, self.port))
            tcp_client.debug(f"[Socket-{self.sock_id}] Connecting to the "\
                             f"Remote Peer")
            self.is_connected = True

            self.selector.register(self.sock, selectors.EVENT_READ | selectors.EVENT_WRITE)
            tcp_client.debug(f"[Socket-{self.sock_id}] Registering Socket "\
                             f"Selector address: {self.selector.get_map()}")

        except Exception as e:
            tcp_client.exception(f"client_errors: {e.args}")


class TcpServer(TcpConnection):
    def __init__(self, ip_address, port):
        super().__init__(ip_address, port)
        

    def start(self):
        try:
            self.server_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            tcp_connection.debug(f"[Socket-{self.sock_id}] Server-side "\
                                 f"Socket: {self.server_sock}")

            self.server_selector = selectors.DefaultSelector()

            self.server_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 4096*64)
            self.server_sock.bind((self.ip_address, self.port))
            self.server_sock.listen()
            tcp_server.debug(f"[Socket-{self.sock_id}] Listening on "\
                             f"{self.ip_address}:{self.port}")

            self.server_sock.setblocking(False)
            tcp_server.debug(f"[Socket-{self.sock_id}] Setting as "\
                             f"Non-Blocking")

            self.server_selector.register(self.server_sock, selectors.EVENT_READ | selectors.EVENT_WRITE)
            tcp_server.debug(f"[Socket-{self.sock_id}] Registering "\
                             f"Socket into Selector address: "\
                             f"{self.server_selector.get_map()}")

        except Exception as e:
            tcp_server.exception(f"server_error: {e.args}")


    def run(self):
        events = self.server_selector.select(timeout=None)
        for key, mask in events:
            tcp_server.debug(f"[Socket-{self.sock_id}] Event has been "\
                             f"raised on Main Socket: (mask, key) = "\
                             f"({mask}, {key})")

            if key.data is None:
                self.sock, self.remote_address = self.server_sock.accept()
                self.sock.setblocking(False)
                tcp_server.debug(f"[Socket-{self.sock_id}] New Socket "\
                                 f"bound to Main Socket: {self.sock}")

                self.is_connected = True
                self.selector.register(self.sock, selectors.EVENT_READ)
                tcp_server.debug(f"[Socket-{self.sock_id}] Registering "\
                                 f"New Socket into Selector address: "\
                                 f"{self.selector.get_map()}")
           
        super().run()


    def close(self):
        super().close()

        try:
            self.server_selector.unregister(self.server_sock)
            tcp_server.debug(f"De-registering Main Socket from Selector "\
                             f"address: {self.server_selector.get_map()}")
    
            self.server_sock.close()
            tcp_server.debug("Shutting down Main Socket")

        except KeyError:
            tcp_server.debug("There is no such Selector registered")



# -*- coding: utf-8 -*-
"""
    bromelia.etsi_3gpp_rx.avps
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains the Diameter protocol AVP library that are used 
    to create Diameter messages for 3GPP Rx Application Id.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from .._internal_utils import show_warn

show_warn("avps", "etsi_3gpp_rx")

from ..avps.etsi_3gpp.ts_129_212 import RatTypeAVP
from ..avps.etsi_3gpp.ts_129_212 import UeLocalIpAddressAVP
from ..avps.etsi_3gpp.ts_129_212 import AnGwAddressAVP
from ..avps.etsi_3gpp.ts_129_212 import IpCanTypeAVP

from ..avps.etsi_3gpp.ts_129_214 import AbortCauseAVP
from ..avps.etsi_3gpp.ts_129_214 import AfApplicationIdentifierAVP
from ..avps.etsi_3gpp.ts_129_214 import AfChargingIdentifierAVP
from ..avps.etsi_3gpp.ts_129_214 import SpecificActionAVP
from ..avps.etsi_3gpp.ts_129_214 import MediaComponentDescriptionAVP
from ..avps.etsi_3gpp.ts_129_214 import ServiceInfoStatusAVP
from ..avps.etsi_3gpp.ts_129_214 import AccessNetworkChargingAddressAVP

from ..avps.etsi_3gpp.ts_129_229 import SupportedFeaturesAVP

from ..avps.etsi_3gpp.ts_129_273 import AnTrustedAVP

from ..avps.etsi_3gpp.ts_183_017 import ReservationPriorityAVP

from ..avps.ietf.rfc4006 import SubscriptionIdAVP

from ..avps.ietf.rfc6733 import AuthApplicationIdAVP
from ..avps.ietf.rfc6733 import AuthSessionStateAVP
from ..avps.ietf.rfc6733 import ClassAVP
from ..avps.ietf.rfc6733 import DestinationHostAVP
from ..avps.ietf.rfc6733 import DestinationRealmAVP
from ..avps.ietf.rfc6733 import ErrorMessageAVP
from ..avps.ietf.rfc6733 import ErrorReportingHostAVP
from ..avps.ietf.rfc6733 import ExperimentalResultAVP
from ..avps.ietf.rfc6733 import FailedAvpAVP
from ..avps.ietf.rfc6733 import OriginHostAVP
from ..avps.ietf.rfc6733 import OriginRealmAVP
from ..avps.ietf.rfc6733 import OriginStateIdAVP
from ..avps.ietf.rfc6733 import ProxyInfoAVP
from ..avps.ietf.rfc6733 import ResultCodeAVP
from ..avps.ietf.rfc6733 import SessionIdAVP
from ..avps.ietf.rfc6733 import RedirectHostAVP
from ..avps.ietf.rfc6733 import RedirectHostUsageAVP
from ..avps.ietf.rfc6733 import RedirectMaxCacheTimeAVP
from ..avps.ietf.rfc6733 import RouteRecordAVP
from ..avps.ietf.rfc6733 import TerminationCauseAVP

from ..avps.ietf.rfc7155 import FramedIpAddressAVP
from ..avps.ietf.rfc7155 import CalledStationIdAVP
from ..avps.ietf.rfc7155 import FramedIpv6PrefixAVP

from ..avps.ietf.rfc8506 import UserEquipmentInfoAVP


# -*- coding: utf-8 -*-
"""
    bromelia.constants.result_codes
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains list of Diameter Result Codes defined in IETF RFC 6733.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""


#: List of DIAMETER Family Codes.

DIAMETER_ERROR_1XXX = bytes.fromhex("000003e8")                 #: 1xxx
DIAMETER_ERROR_2XXX = bytes.fromhex("000007d0")                 #: 2xxx
DIAMETER_ERROR_3XXX = bytes.fromhex("00000bb8")                 #: 3xxx
DIAMETER_ERROR_4XXX = bytes.fromhex("00000fa0")                 #: 4xxx
DIAMETER_ERROR_5XXX = bytes.fromhex("00001388")                 #: 5xxx

#: Informational Result-Codes.
DIAMETER_MULTI_ROUND_AUTH = bytes.fromhex("000003e9")           #: 1001 - Subsequent messages triggered by client shall also used in Authentication and to get access of required resources. Generally used in Diameter NAS.

#: Success Result-Codes.
DIAMETER_SUCCESS = bytes.fromhex("000007d1")                    #: 2001 - Request processed successfully.
DIAMETER_LIMITED_SUCCESS = bytes.fromhex("000007d2")            #: 2002 - Request is processed but some more processing is required by Server to provide access to user.

#: Protocol Errors [E-bit set].
DIAMETER_COMMAND_UNSUPPORTED = bytes.fromhex("00000bb9")        #: 3001 - Server returns it if Diameter Command-Code is un-recognized by server.
DIAMETER_UNABLE_TO_DELIVER = bytes.fromhex("00000bba")          #: 3002 - Message can’t be delivered because there is no Host with Diameter URI present in Destination-Host AVP in associated Realm.
DIAMETER_REALM_NOT_SERVED = bytes.fromhex("00000bbb")           #: 3003 - Intended Realm is not recognized.
DIAMETER_TOO_BUSY = bytes.fromhex("00000bbc")                   #: 3004 - Shall return by server only when server unable to provide requested service, where all the pre-requisites are also met. Client should also send the request to alternate peer.
DIAMETER_LOOP_DETECTED = bytes.fromhex("00000bbd")              #: 3005
DIAMETER_REDIRECT_INDICATION = bytes.fromhex("00000bbe")        #: 3006 - In Response from Redirect Agent.
DIAMETER_APPLICATION_UNSUPPORTED = bytes.fromhex("00000bbf")    #: 3007
DIAMETER_INVALID_HDR_BITS = bytes.fromhex("00000bc0")           #: 3008 - It is sent when a request is received with invalid bits combination for considered command-code in DIAMETER Header structure. E.g. Marking Proxy-Bit in CER message.
DIAMETER_INVALID_AVP_BITS = bytes.fromhex("00000bc1")           #: 3009 - It is sent when a request is received with invalid flag bits in an AVP.
DIAMETER_UNKNOWN_PEER = bytes.fromhex("00000bc2")               #: 3010 - A DIAMETER server can be configured whether it shall accept DIAMETER connection from all nodes or only from specific nodes. If it is configured to accept connection from specific nodes and receives CER from message from any node other than specified. Here Server shall send considered error.

#: Transient Failures [Could not satisfy request at this moment].
DIAMETER_AUTHENTICATION_REJECTED = bytes.fromhex("00000fa1")    #: 4001 - Returned by Server, most likely because of invalid password.
DIAMETER_OUT_OF_SPACE = bytes.fromhex("00000fa2")               #: 4002 - Returned by node, when it receives accounting information but unable to store it because of lack of memory.
DIAMETER_ELECTION_LOST = bytes.fromhex("00000fa3")              #: 4003 - Peer determines that it has lost election by comparing Origin-Host value received in CER with its own DIAMETER IDENTITY and found that received DIAMETER IDENTITY is higher.

#: Permanent Failures [To inform peer, request is failed, shouldn’t be attempted again].
DIAMETER_AVP_UNSUPPORTED = bytes.fromhex("00001389")            #: 5001 - AVP marked with Mandatory Bit, but peer does not support it.
DIAMETER_UNKNOWN_SESSION_ID = bytes.fromhex("0000138a")         #: 5002
DIAMETER_AUTHORIZATION_REJECTED = bytes.fromhex("0000138b")     #: 5003 - User can not be authorized. E.g. Comes in AIA on s6a interface.
DIAMETER_INVALID_AVP_VALUE = bytes.fromhex("0000138c")          #: 5004
DIAMETER_MISSING_AVP = bytes.fromhex("0000138d")                #: 5005 - Mandatory AVP in request message is missing.
DIAMETER_RESOURCES_EXCEEDED = bytes.fromhex("0000138e")         #: 5006 - A request was received that cannot be authorized because the user has already expended allowed resources. An example of this error condition is a user that is restricted to one dial-up PPP port, attempts to establish a second PPP connection.
DIAMETER_CONTRADICTING_AVPS = bytes.fromhex("0000138f")         #: 5007 - Server has identified that AVPs are present that are contradictory to each other.
DIAMETER_AVP_NOT_ALLOWED = bytes.fromhex("00001390")            #: 5008 - Message is received by node (Server) that contain AVP must not be present.
DIAMETER_AVP_OCCURS_TOO_MANY_TIMES = bytes.fromhex("00001391")  #: 5009 - If message contains the a AVP number of times that exceeds permitted occurrence of AVP in message definition.
DIAMETER_NO_COMMON_APPLICATION = bytes.fromhex("00001392")      #: 5010 - In response of CER if no common application supported between the peers.
DIAMETER_UNSUPPORTED_VERSION = bytes.fromhex("00001393")        #: 5011 - Self explanatory.
DIAMETER_UNABLE_TO_COMPLY = bytes.fromhex("00001394")           #: 5012 - Message rejected because of unspecified reasons.
DIAMETER_INVALID_BIT_IN_HEADER = bytes.fromhex("00001395")      #: 5013 - When an unrecognized bit in the Diameter header is set to one.
DIAMETER_INVALID_AVP_LENGTH = bytes.fromhex("00001396")         #: 5014 - Self explanatory. 
DIAMETER_INVALID_MESSAGE_LENGTH = bytes.fromhex("00001397")     #: 5015 - Self explanatory.
DIAMETER_INVALID_AVP_BIT_COMBO = bytes.fromhex("00001398")      #: 5016 - E.g. Marking AVP to Mandatory while message definition doesn’t say so.
DIAMETER_NO_COMMON_SECURITY = bytes.fromhex("00001399")         #: 5017 - In response of CER if no common security mechanism supported between the peers.
